import copy
import io
import os.path
import pkgutil
from collections import namedtuple
from itertools import groupby

import pyte.graphics
import pyte.screens
from lxml import etree
from wcwidth import wcswidth

# Ugliest hack: Replace the first 16 colors rgb values by their names so that
# termtosvg can distinguish FG_BG_256[0] (which defaults to black #000000 but
# can be styled with themes) from FG_BG_256[16] (which is also black #000000
# but should be displayed as is).
_COLORS = ['black', 'red', 'green', 'brown', 'blue', 'magenta', 'cyan', 'white']
_BRIGHTCOLORS = ['bright{}'.format(color) for color in _COLORS]
NAMED_COLORS = _COLORS + _BRIGHTCOLORS
pyte.graphics.FG_BG_256 = NAMED_COLORS + pyte.graphics.FG_BG_256[16:]

# Id for the very last SVG animation. This is used to make the first animations
# start when the last one ends (animation looping)
LAST_ANIMATION_ID = 'anim_last'

# Background rectangle SVG element
_BG_RECT_TAG_ATTRIBUTES = {
    'class': 'background',
    'height': '100%',
    'width': '100%',
    'x': '0',
    'y': '0'
}
BG_RECT_TAG = etree.Element('rect', _BG_RECT_TAG_ATTRIBUTES)

# Default size for a character cell rendered as SVG.
CELL_WIDTH = 8
CELL_HEIGHT = 17

# The number of character cells to leave when placing successive frames
# so content does not bleed into adjacent frames
FRAME_CELL_SPACING = 1

# XML namespaces
SVG_NS = 'http://www.w3.org/2000/svg'
TERMTOSVG_NS = 'https://github.com/nbedos/termtosvg'
XLINK_NS = 'http://www.w3.org/1999/xlink'

NAMESPACES = {
    'svg': SVG_NS,
    'termtosvg': TERMTOSVG_NS,
    'xlink': XLINK_NS,
}


class TemplateError(Exception):
    pass


_CELL_ATTRIBUTES = ['text', 'color', 'background_color', 'bold', 'italics',
                    'underscore', 'strikethrough']
_CharacterCell = namedtuple('_CharacterCell', _CELL_ATTRIBUTES)
# Set default values for last 6 arguments
_CharacterCell.__new__.__defaults__ = ('foreground', 'background', False, False,
                                       False, False)
_CharacterCell.__doc__ = 'Representation of a character cell'
_CharacterCell.text.__doc__ = 'Text content of the cell'
_CharacterCell.bold.__doc__ = 'Bold modificator flag'
_CharacterCell.italics.__doc__ = 'Italics modificator flag'
_CharacterCell.underscore.__doc__ = 'Underscore modificator flag'
_CharacterCell.strikethrough.__doc__ = 'Strikethrough modificator flag'
_CharacterCell.color.__doc__ = 'Color of the text'
_CharacterCell.background_color.__doc__ = 'Background color of the cell'


class CharacterCell(_CharacterCell):
    @classmethod
    def from_pyte(cls, char):
        """Create a CharacterCell from a pyte character"""
        if char.fg == 'default':
            text_color = 'foreground'
        else:
            if char.bold and not str(char.fg).startswith('bright'):
                named_color = 'bright{}'.format(char.fg)
            else:
                named_color = char.fg

            if named_color in NAMED_COLORS:
                text_color = 'color{}'.format(NAMED_COLORS.index(named_color))
            elif len(char.fg) == 6:
                # HEXADECIMAL COLORS
                # raise ValueError if char.fg is not an hexadecimal number
                int(char.fg, 16)
                text_color = '#{}'.format(char.fg)
            else:
                raise ValueError('Invalid foreground color: {}'.format(char.fg))

        if char.bg == 'default':
            background_color = 'background'
        elif char.bg in NAMED_COLORS:
            background_color = 'color{}'.format(NAMED_COLORS.index(char.bg))
        elif len(char.bg) == 6:
            # Hexadecimal colors
            # raise ValueError if char.bg is not an hexadecimal number
            int(char.bg, 16)
            background_color = '#{}'.format(char.bg)
        else:
            raise ValueError('Invalid background color')

        if char.reverse:
            text_color, background_color = background_color, text_color

        return CharacterCell(char.data, text_color, background_color,
                             char.bold, char.italics, char.underscore,
                             char.strikethrough)


class ConsecutiveWithSameAttributes:
    """Callable to be used as a key for itertools.groupby to group together
    consecutive elements of a list with the same attributes"""
    def __init__(self, attributes):
        self.group_index = None
        self.last_index = None
        self.attributes = attributes
        self.last_key_attributes = None

    def __call__(self, arg):
        index, obj = arg
        key_attributes = {name: getattr(obj, name) for name in self.attributes}
        if self.last_index != index - 1 or self.last_key_attributes != key_attributes:
            self.group_index = index
        self.last_index = index
        self.last_key_attributes = key_attributes
        return self.group_index, key_attributes


def render_animation(frames, geometry, filename, template,
                     cell_width=CELL_WIDTH, cell_height=CELL_HEIGHT):
    root = _render_preparation(geometry, template, cell_width, cell_height)
    _, screen_height = geometry
    root = _render_animation(screen_height, frames, root, cell_width, cell_height)

    with open(filename, 'wb') as output_file:
        output_file.write(etree.tostring(root))


def render_still_frames(frames, geometry, directory, template,
                        cell_width=CELL_WIDTH, cell_height=CELL_HEIGHT):
    root = _render_preparation(geometry, template, cell_width, cell_height)

    frame_generator = _render_still_frames(frames, root, cell_width, cell_height)
    for frame_count, frame_root in enumerate(frame_generator):
        filename = os.path.join(directory, 'termtosvg_{:05}.svg'.format(frame_count))
        with open(filename, 'wb') as output_file:
            output_file.write(etree.tostring(frame_root))


def _render_preparation(geometry, template, cell_width, cell_height):
    # Read header record and add the corresponding information to the SVG
    root = resize_template(template, geometry, cell_width, cell_height)
    svg_screen_tag = root.find('.//{{{namespace}}}svg[@id="screen"]'
                               .format(namespace=SVG_NS))
    if svg_screen_tag is None:
        raise ValueError('Missing tag: <svg id="screen" ...>...</svg>')

    for child in svg_screen_tag.getchildren():
        svg_screen_tag.remove(child)
    svg_screen_tag.append(BG_RECT_TAG)

    return root


def _render_still_frames(frames, root, cell_width, cell_height):
    for frame in frames:
        frame_group, frame_definitions = _render_timed_frame(
            offset=0,
            buffer=frame.buffer,
            cell_height=cell_height,
            cell_width=cell_width,
            definitions={}
        )
        frame_root = copy.deepcopy(root)
        svg_screen_tag = frame_root.find('.//{{{namespace}}}svg[@id="screen"]'
                                         .format(namespace=SVG_NS))
        if svg_screen_tag is None:
            raise ValueError('Missing tag: <svg id="screen" ...>...</svg>')
        tree_defs = etree.SubElement(svg_screen_tag, 'defs')
        for definition in frame_definitions.values():
            tree_defs.append(definition)
        svg_screen_tag.append(frame_group)

        _embed_css(frame_root)
        yield frame_root


def _render_animation(screen_height, frames, root, cell_width, cell_height):
    svg_screen_tag = root.find('.//{{{namespace}}}svg[@id="screen"]'
                               .format(namespace=SVG_NS))
    if svg_screen_tag is None:
        raise ValueError('Missing tag: <svg id="screen" ...>...</svg>')

    screen_view = etree.Element('g', attrib={'id': 'screen_view'})

    definitions = {}
    timings = {}
    animation_duration = None
    for frame_count, frame in enumerate(frames):
        # To prevent line jumping up and down by one pixel between two frames,
        # add h % 2 so that offset is an even number.  (issue noticed in
        # Firefox only)
        h = screen_height + FRAME_CELL_SPACING
        offset = frame_count * (h + h % 2) * cell_height
        frame_group, frame_definitions = _render_timed_frame(
            offset=offset,
            buffer=frame.buffer,
            cell_height=cell_height,
            cell_width=cell_width,
            definitions=definitions
        )

        screen_view.append(frame_group)
        animation_duration = frame.time + frame.duration
        timings[frame.time] = -offset
        definitions.update(frame_definitions)

    tree_defs = etree.SubElement(svg_screen_tag, 'defs')
    for definition in definitions.values():
        tree_defs.append(definition)

    svg_screen_tag.append(screen_view)
    _add_animation(root, timings, animation_duration)
    return root


def _add_animation(root, timings, animation_duration):
    animators = {
        'css': _embed_css,
        'waapi': _embed_waapi,
    }

    settings = root.find('.//{{{}}}defs/{{{}}}template_settings'
                         .format(SVG_NS, TERMTOSVG_NS))
    if settings is None:
        raise TemplateError('Missing "template_settings" element in definitions')

    animation = settings.find('{{{}}}animation[@type]'.format(TERMTOSVG_NS))
    if animation is None:
        raise TemplateError('Missing or invalid "animation" element in "template_settings"')

    f = animators.get(animation.attrib['type'].lower())
    if f is None:
        raise TemplateError("Attribute 'type' of element 'animation' must be one of {}"
                            .format(', '.join(animators.keys())))

    f(root, timings, animation_duration)


def _render_timed_frame(offset, buffer, cell_height, cell_width, definitions):
    """Return a group element containing an SVG version of the provided frame.

    :param buffer: 2D array of CharacterCells
    :param cell_height: Height of a character cell in pixels
    :param cell_width: Width of a character cell in pixels
    :param definitions: Existing definitions (updated in place)
    :return: A tuple consisting of a group element and new definitions
    """
    frame_group_tag = etree.Element('g')

    group_definitions = {}
    for row_number in buffer:
        if buffer[row_number]:
            current_definitions = {**definitions, **group_definitions}
            tags, new_definitions = _render_line(offset,
                                                 row_number,
                                                 buffer[row_number],
                                                 cell_height,
                                                 cell_width,
                                                 current_definitions)
            for tag in tags:
                frame_group_tag.append(tag)
            group_definitions.update(new_definitions)

    return frame_group_tag, group_definitions


def _render_line(offset, row_number, row, cell_height, cell_width, definitions):
    tags = _render_line_bg_colors(screen_line=row,
                                  height=offset + row_number * cell_height,
                                  cell_height=cell_height,
                                  cell_width=cell_width)

    # Group text elements for the current line into text_group_tag
    text_group_tag = etree.Element('g')
    text_tags = _render_characters(row, cell_width)
    for tag in text_tags:
        text_group_tag.append(tag)

    # Find or create a definition for text_group_tag
    text_group_tag_str = etree.tostring(text_group_tag)
    if text_group_tag_str in definitions:
        group_id = definitions[text_group_tag_str].attrib['id']
        new_definitions = {}
    else:
        group_id = 'g{}'.format(len(definitions) + 1)
        assert group_id not in definitions.values()
        text_group_tag.attrib['id'] = group_id
        new_definitions = {text_group_tag_str: text_group_tag}

    # Add a reference to the definition of text_group_tag with a 'use' tag
    use_attributes = {
        '{{{}}}href'.format(XLINK_NS): '#{}'.format(group_id),
        'y': str(offset + row_number * cell_height),
    }
    tags.append(etree.Element('use', use_attributes))

    return tags, new_definitions


def _make_rect_tag(column, length, height, cell_width, cell_height, background_color):
    attributes = {
        'x': str(column * cell_width),
        'y': str(height),
        'width': str(length * cell_width),
        'height': str(cell_height)
    }

    if background_color.startswith('#'):
        attributes['fill'] = background_color
    else:
        attributes['class'] = background_color
    rect_tag = etree.Element('rect', attributes)
    return rect_tag


def _render_line_bg_colors(screen_line, height, cell_height, cell_width):
    """Return a list of 'rect' tags representing the background of 'screen_line'

    If consecutive cells have the same background color, a single 'rect' tag is
    returned for all these cells.
    If a cell background uses default_bg_color, no 'rect' will be generated for
    this cell since the default background is always displayed.

    :param screen_line: Mapping between column numbers and CharacterCells
    :param height: Vertical position of the line on the screen in pixels
    :param cell_height: Height of the a character cell in pixels
    :param cell_width: Width of a character cell in pixels
    """
    non_default_bg_cells = [(column, cell) for (column, cell)
                            in sorted(screen_line.items())
                            if cell.background_color != 'background']

    key = ConsecutiveWithSameAttributes(['background_color'])
    rect_tags = [
        _make_rect_tag(
            column,
            wcswidth(''.join(t[1].text for t in group)),
            height,
            cell_width,
            cell_height,
            attributes['background_color']
        ) for (column, attributes), group in groupby(non_default_bg_cells, key)]

    return rect_tags


def _make_text_tag(column, attributes, text, cell_width):
    """Build SVG text element based on content and style attributes"""
    text_tag_attributes = {
        'x': str(column * cell_width),
        'textLength': str(wcswidth(text) * cell_width),
    }
    if attributes['bold']:
        text_tag_attributes['font-weight'] = 'bold'

    if attributes['italics']:
        text_tag_attributes['font-style'] = 'italic'

    decoration = ''
    if attributes['underscore']:
        decoration = 'underline'
    if attributes['strikethrough']:
        decoration += ' line-through'
    if decoration:
        text_tag_attributes['text-decoration'] = decoration

    if attributes['color'].startswith('#'):
        text_tag_attributes['fill'] = attributes['color']
    else:
        text_tag_attributes['class'] = attributes['color']

    text_tag = etree.Element('text', text_tag_attributes)
    text_tag.text = text
    return text_tag


def _render_characters(screen_line, cell_width):
    """Return a list of 'text' elements representing the line of the screen

    Consecutive characters with the same styling attributes (text color, font
    weight...) are grouped together in a single text element.

    :param screen_line: Mapping between column numbers and characters
    :param cell_width: Width of a character cell in pixels
    """
    line = sorted(screen_line.items())
    key = ConsecutiveWithSameAttributes(['color', 'bold', 'italics', 'underscore', 'strikethrough'])
    text_tags = [_make_text_tag(column, attributes, ''.join(c.text for _, c in group), cell_width)
                 for (column, attributes), group in groupby(line, key)]

    return text_tags


def resize_template(template, geometry, cell_width, cell_height):
    """Resize template based on the number of rows and columns of the terminal"""
    def scale(element, template_columns, template_rows, columns, rows):
        """Resize viewbox based on the number of rows and columns of the terminal"""
        try:
            viewbox = element.attrib['viewBox'].replace(',', ' ').split()
        except KeyError:
            raise TemplateError('Missing "viewBox" for element "{}"'.format(element))

        vb_min_x, vb_min_y, vb_width, vb_height = [int(n) for n in viewbox]
        vb_width += cell_width * (columns - template_columns)
        vb_height += cell_height * (rows - template_rows)
        element.attrib['viewBox'] = ' '.join(map(str, (vb_min_x, vb_min_y, vb_width, vb_height)))

        scalable_attributes = {
            'width': cell_width * (columns - template_columns),
            'height': cell_height * (rows - template_rows)
        }

        for attribute, delta in scalable_attributes.items():
            if attribute in element.attrib:
                try:
                    element.attrib[attribute] = str(int(element.attrib[attribute]) + delta)
                except ValueError:
                    raise TemplateError('"{}" attribute of {} must be in user units'
                                        .format(attribute, element))
        return element

    try:
        tree = etree.parse(io.BytesIO(template))
        root = tree.getroot()
    except etree.Error as exc:
        raise TemplateError('Invalid template') from exc

    # Extract the screen geometry which is saved in a private data portion of
    # the template
    settings = root.find('.//{{{}}}defs/{{{}}}template_settings'
                         .format(SVG_NS, TERMTOSVG_NS))
    if settings is None:
        raise TemplateError('Missing "template_settings" element in definitions')

    svg_geometry = settings.find('{{{}}}screen_geometry[@columns][@rows]'
                                 .format(TERMTOSVG_NS))
    if svg_geometry is None:
        raise TemplateError('Missing "screen_geometry" element in "template_settings"')

    attributes_err_msg = ('Missing or invalid "columns" or "rows" attribute '
                          'for element "screen_geometry": expected positive '
                          'integers')
    try:
        template_columns = int(svg_geometry.attrib['columns'])
        template_rows = int(svg_geometry.attrib['rows'])
    except (KeyError, ValueError) as exc:
        raise TemplateError(attributes_err_msg) from exc

    # Update settings with real columns and rows values to preserve the scale
    # in case the animation serves as a template
    columns, rows = geometry
    svg_geometry.attrib['columns'], svg_geometry.attrib['rows'] = (str(columns),
                                                                   str(rows))

    if template_rows <= 0 or template_columns <= 0:
        raise TemplateError(attributes_err_msg)

    # Scale the viewBox of the root svg element based on the size of the screen
    # and the size registered in the template
    scale(root, template_columns, template_rows, columns, rows)

    # Also scale the viewBox of the svg element with id 'screen'
    screen = root.find('.//{{{namespace}}}svg[@id="screen"]'
                       .format(namespace=SVG_NS))
    if screen is None:
        raise TemplateError('svg element with id "screen" not found')
    scale(screen, template_columns, template_rows, columns, rows)

    return root


def validate_template(name, templates):
    if name in templates:
        return templates[name]

    try:
        with open(name, 'rb') as template_file:
            return template_file.read()
    except FileNotFoundError as exc:
        raise TemplateError('Invalid template') from exc


def _embed_css(root, timings=None, animation_duration=None):
    try:
        style = root.find('.//{{{ns}}}defs/{{{ns}}}style[@id="generated-style"]'
                          .format(ns=SVG_NS))
    except etree.Error as exc:
        raise TemplateError('Invalid template') from exc

    if style is None:
        raise TemplateError('Missing <style id="generated-style" ...> element '
                            'in "defs"')

    css_body = """#screen {
                font-family: 'DejaVu Sans Mono', monospace;
                font-style: normal;
                font-size: 14px;
            }

        text {
            dominant-baseline: text-before-edge;
            white-space: pre;
        }
    """

    if animation_duration is None or timings is None:
        style.text = etree.CDATA(css_body)
    else:
        if animation_duration == 0:
            raise ValueError('Animation duration must be greater than 0')

        transforms = []
        last_offset = None
        transform_format = "{time:.3f}%{{transform:translateY({offset}px)}}"
        for time, offset in sorted(timings.items()):
            transforms.append(
                transform_format.format(
                    time=100.0 * time/animation_duration,
                    offset=offset
                )
            )
            last_offset = offset

        if last_offset is not None:
            transforms.append(
                transform_format.format(time=100, offset=last_offset)
            )

        css_animation = """
            :root {{
                --animation-duration: {duration}ms;
            }}

            @keyframes roll {{
                {transforms}
            }}

            #screen_view {{
                animation-duration: {duration}ms;
                animation-iteration-count:infinite;
                animation-name:roll;
                animation-timing-function: steps(1,end);
                animation-fill-mode: forwards;
            }}
        """.format(
            duration=animation_duration,
            transforms=os.linesep.join(transforms)
        )

        style.text = etree.CDATA(css_body + css_animation)

    return root


def _embed_waapi(root, timings=None, animation_duration=None):
    try:
        style = root.find('.//{{{ns}}}defs/{{{ns}}}style[@id="generated-style"]'
                          .format(ns=SVG_NS))
    except etree.Error as exc:
        raise TemplateError('Invalid template') from exc

    if style is None:
        raise TemplateError('Missing <style id="generated-style" ...> element '
                            'in "defs"')

    css_body = """
        #screen {
            font-family: 'DejaVu Sans Mono', monospace;
            font-style: normal;
            font-size: 14px;
        }

        text {
            dominant-baseline: text-before-edge;
            white-space: pre;
        }
    """

    style.text = etree.CDATA(css_body)

    if animation_duration and timings:
        if animation_duration == 0:
            raise ValueError('Animation duration must be greater than 0')

        css_body += """
        :root {{
            --animation-duration: {duration}ms;
        }}

        """.format(duration=animation_duration)

        script_element = root.find('.//{{{ns}}}script[@id="generated-js"]'
                                   .format(ns=SVG_NS))
        if script_element is None:
            raise TemplateError(
                'Missing <script id="generated-js" ...> element')


        transform_no_offset = "{{transform: 'translate3D(0, {y_pos}px, 0)', easing: 'steps(1, end)'}}"
        transform_with_offset = "{{transform: 'translate3D(0, {y_pos}px, 0)', easing: 'steps(1, end)', offset: {offset:.3f}}}"

        transforms = []
        last_pos = None
        for time, y_pos in sorted(timings.items()):
            if last_pos is None:
                transforms.append(transform_no_offset.format(y_pos=y_pos))
            else:
                transforms.append(
                    transform_with_offset
                    .format(offset=time / animation_duration, y_pos=y_pos)
                )
            last_pos = y_pos

        if last_pos is not None:
            transforms.append(transform_no_offset.format(y_pos=last_pos))

        js_animation = """
        var termtosvg_vars = {{
            transforms: [
                {transforms}
            ],
            timings: {{
                duration: {duration},
                iterations: Infinity
            }}
        }};""".format(
            transforms=',{}'.format(os.linesep).join(transforms),
            duration=animation_duration
        )

        script_element.text = etree.CDATA(js_animation)

    return root


def validate_svg(svg_file):
    """Validate an SVG file against SVG 1.1 Document Type Definition"""
    package = __name__.split('.')[0]
    dtd_bytes = pkgutil.get_data(package, '/data/svg11-flat-20110816.dtd')
    with io.BytesIO(dtd_bytes) as bstream:
        dtd = etree.DTD(bstream)

    try:
        tree = etree.parse(svg_file)
        for bad in tree.xpath('/svg:svg/svg:defs/termtosvg:template_settings',
                              namespaces=NAMESPACES):
            bad.getparent().remove(bad)
        root = tree.getroot()
        is_valid = dtd.validate(root)
    except etree.Error as exc:
        raise ValueError('Invalid SVG file') from exc

    if not is_valid:
        reason = dtd.error_log.filter_from_errors()[0]
        raise ValueError('Invalid SVG file: {}'.format(reason))



import pkgutil

PKG_TEMPLATE_PATH = 'data/templates'

# Listing templates here is not ideal but importing pkg_resources to get resource_listdir
# does not seem worth it: it adds a dependency and slows down the invocation of termtosvg
# by 150ms ('time termtosvg --help' execution time goes from 200ms to 350ms)
DEFAULT_TEMPLATES_NAMES = [
    'base16_default_dark.svg',
    'dracula.svg',
    'gjm8_play.svg',
    'gjm8_single_loop.svg',
    'gjm8.svg',
    'powershell.svg',
    'progress_bar.svg',
    'putty.svg',
    'solarized_dark.svg',
    'solarized_light.svg',
    'terminal_app.svg',
    'ubuntu.svg',
    'window_frame_js.svg',
    'window_frame_powershell.svg',
    'window_frame.svg',
    'xterm.svg',
]


def validate_geometry(screen_geometry):
    """Raise ValueError if 'screen_geometry' does not conform to <integer>x<integer> format"""
    columns, rows = [int(value) for value in screen_geometry.lower().split('x')]
    if columns <= 0 or rows <= 0:
        raise ValueError('Invalid value for screen-geometry option: "{}"'.format(screen_geometry))
    return columns, rows


def default_templates():
    """Return mapping between the name of a template and the SVG template itself"""
    templates = {}
    for template_name in DEFAULT_TEMPLATES_NAMES:
        pkg_template_path = '{}/{}'.format(PKG_TEMPLATE_PATH, template_name)
        bstream = pkgutil.get_data(__name__, pkg_template_path)
        suffix = '.svg'
        if template_name.endswith(suffix):
            templates[template_name[:-len(suffix)]] = bstream
        else:
            templates[template_name] = bstream

    return templates



"""asciicast records

This module provides functions and classes to manipulate asciicast records. Both v1 and v2
format are supported for decoding. For encoding, only the v2 format is available. The
specification of both formats are available here:
    [1] https://github.com/asciinema/asciinema/blob/develop/doc/asciicast-v1.md
    [2] https://github.com/asciinema/asciinema/blob/develop/doc/asciicast-v2.md
"""
import abc
import codecs
import json
from collections import namedtuple
from typing import Iterable


class AsciiCastError(Exception):
    pass


class AsciiCastV2Record(abc.ABC):
    """Generic Asciicast v2 record format"""
    @abc.abstractmethod
    def to_json_line(self):
        raise NotImplementedError

    @classmethod
    def from_json_line(cls, line):
        """Raise AsciiCastError if line is not a valid asciicast v2 record"""
        try:
            json_dict = json.loads(line)
        except json.JSONDecodeError as exc:
            raise AsciiCastError from exc
        if isinstance(json_dict, dict):
            return AsciiCastV2Header.from_json_line(line)
        if isinstance(json_dict, list):
            return AsciiCastV2Event.from_json_line(line)
        truncated_line = line if len(line) < 20 else '{}...'.format(line[:20])
        raise AsciiCastError('Unknown record type: "{}"'.format(truncated_line))


def _read_v1_records(data):
    v1_header_attributes = {
        'version',
        'width',
        'height',
        'stdout'
    }
    try:
        json_dict = json.loads(data)
    except json.JSONDecodeError as exc:
        raise AsciiCastError from exc
    missing_attributes = v1_header_attributes - set(json_dict)
    if missing_attributes:
        raise AsciiCastError('Missing attributes in asciicast v1 file: {}'
                             .format(missing_attributes))

    if json_dict['version'] != 1:
        raise AsciiCastError('This function can only decode asciicast v1 data')

    yield AsciiCastV2Header(2, json_dict['width'], json_dict['height'], None, None)

    if not isinstance(json_dict['stdout'], Iterable):
        raise AsciiCastError('Invalid type for stdout attribute (expected Iterable): {}'
                             .format(json_dict['stdout']))

    time = 0
    for event in json_dict['stdout']:
        try:
            time_elapsed, event_data = event
        except ValueError as exc:
            raise AsciiCastError from exc

        if not isinstance(time_elapsed, (int, float)) or not isinstance(event_data, str):
            raise AsciiCastError('Invalid type for event: got object "{}" but expected '
                                 'type Tuple[Union[int, float], str]'.format(event))
        time += time_elapsed
        yield AsciiCastV2Event(time, 'o', event_data, None)


def read_records(filename):
    """Yield asciicast v2 records from the file

    The records in the file may themselves be in either asciicast v1 or v2 format (although
    there must be only one record format version in the file).
    Raise AsciiCastError if a record is invalid"""
    try:
        with open(filename, 'r') as cast_file:
            for line in cast_file:
                yield AsciiCastV2Record.from_json_line(line)
    except AsciiCastError:
        with open(filename, 'r') as cast_file:
            yield from _read_v1_records(cast_file.read())


_AsciiCastV2Theme = namedtuple('AsciiCastV2Theme', ['fg', 'bg', 'palette'])


class AsciiCastV2Theme(_AsciiCastV2Theme):
    """Color theme of the terminal.

    All colors must use the '#rrggbb' format

    fg: default text color
    bg: default background colors
    palette: colon separated list of 8 or 16 terminal colors
    """
    def __new__(cls, fg, bg, palette):
        if cls.is_color(fg):
            if cls.is_color(bg):
                colors = palette.split(':')
                if len(colors) >= 16 and all([cls.is_color(c) for c in colors[:16]]):
                    self = super().__new__(cls, fg, bg, palette)
                    return self
                if len(colors) >= 8 and all([cls.is_color(c) for c in colors[:8]]):
                    new_palette = ':'.join(colors[:8])
                    self = super().__new__(cls, fg, bg, new_palette)
                    return self
                raise AsciiCastError('Invalid palette: the first 8 or 16 colors must be valid')
            raise AsciiCastError('Invalid background color: {}'.format(bg))
        raise AsciiCastError('Invalid foreground color: {}'.format(fg))

    @staticmethod
    def is_color(color):
        if isinstance(color, str) and len(color) == 7 and color[0] == '#':
            try:
                int(color[1:], 16)
            except ValueError:
                return False
            return True
        return False


_AsciiCastV2Header = namedtuple('AsciiCastV2Header', ['version', 'width', 'height', 'theme',
                                                      'idle_time_limit'])


class AsciiCastV2Header(AsciiCastV2Record, _AsciiCastV2Header):
    """Header record

    version: Version of the asciicast file format
    width: Initial number of columns of the terminal
    height: Initial number of lines of the terminal
    theme: Color theme of the terminal
    """
    types = {
        'version': int,
        'width': int,
        'height': int,
        'theme': (type(None), AsciiCastV2Theme),
        'idle_time_limit': (type(None), int, float)
    }

    def __new__(cls, version, width, height, theme, idle_time_limit=None):
        self = super(AsciiCastV2Header, cls).__new__(cls, version, width, height, theme, idle_time_limit)
        for attr_name in cls._fields:
            attr = self.__getattribute__(attr_name)
            if not isinstance(attr, cls.types[attr_name]):
                raise AsciiCastError('Invalid type for attribute {}: {} (expected one of {})'
                                     .format(attr_name, type(attr), cls.types[attr_name]))
        if version != 2:
            raise AsciiCastError('Only asciicast v2 format is supported')
        return self

    def to_json_line(self):
        attributes = self._asdict()
        if self.theme is not None:
            attributes['theme'] = self.theme._asdict()
        else:
            del attributes['theme']

        if attributes['idle_time_limit'] is None:
            del attributes['idle_time_limit']

        return json.dumps(attributes, ensure_ascii=False)

    @classmethod
    def from_json_line(cls, line):
        attributes = json.loads(line)
        filtered_attributes = {attr: attributes.get(attr) for attr in AsciiCastV2Header._fields}
        if filtered_attributes['theme'] is not None:
            filtered_attributes['theme'] = AsciiCastV2Theme(**filtered_attributes['theme'])

        header = cls(**filtered_attributes)
        return header


_AsciiCastV2Event = namedtuple('AsciiCastV2Event', ['time', 'event_type', 'event_data', 'duration'])


class AsciiCastV2Event(AsciiCastV2Record, _AsciiCastV2Event):
    """Event record

    time: Time elapsed since the beginning of the recording in seconds
    event_type: Type 'o' if the data was captured on the standard output of the terminal, type
                'i' if it was captured on the standard input
    event_data: Data captured during the recording
    duration: Duration of the event in seconds (non standard field)
    """
    types = {
        'time': (int, float),
        'event_type': (str,),
        'event_data': (str,),
        'duration': (type(None), int, float),
    }

    def __new__(cls, *args, **kwargs):
        self = super(AsciiCastV2Event, cls).__new__(cls, *args, **kwargs)
        for attr_name in AsciiCastV2Event._fields:
            attr = self.__getattribute__(attr_name)
            if not isinstance(attr, cls.types[attr_name]):
                raise AsciiCastError('Invalid type for attribute {}: {} (expected one of {})'
                                     .format(attr_name, type(attr), cls.types[attr_name]))
        return self

    def to_json_line(self):
        attributes = [self.time, self.event_type, self.event_data]
        return json.dumps(attributes, ensure_ascii=False)

    @classmethod
    def from_json_line(cls, line):
        try:
            time, event_type, event_data = json.loads(line)
        except (json.JSONDecodeError, ValueError) as exc:
            raise AsciiCastError from exc

        event = cls(time, event_type, event_data, None)
        return event



"""Command line interface of termtosvg"""

import argparse
import logging
import os
import shlex
import sys
import tempfile

import termtosvg.config
import termtosvg.anim

logger = logging.getLogger('termtosvg')

DEFAULT_LOOP_DELAY = 1000

USAGE = """termtosvg [output_path] [-c COMMAND] [-D DELAY] [-g GEOMETRY]
                 [-m MIN_DURATION] [-M MAX_DURATION] [-s] [-t TEMPLATE] [-h]

Record a terminal session and render an SVG animation on the fly
"""
EPILOG = "See also 'termtosvg record --help' and 'termtosvg render --help'"
RECORD_USAGE = "termtosvg record [output_path] [-c COMMAND] [-g GEOMETRY] [-h]"
RENDER_USAGE = """termtosvg render input_file [output_path] [-D DELAY]
                 [-m MIN_DURATION] [-M MAX_DURATION] [-s] [-t TEMPLATE] [-h]"""


def integral_duration_validation(duration):
    if duration.lower().endswith('ms'):
        duration = duration[:-len('ms')]

    if duration.isdigit() and int(duration) >= 1:
        return int(duration)
    raise ValueError('duration must be an integer greater than 0')


def parse(args, templates, default_template, default_geometry, default_min_dur,
          default_max_dur, default_cmd, default_loop_delay):
    """Parse command line arguments

    :param args: Arguments to parse
    :param templates: Mapping between template names and templates
    :param default_template: Name of the default template
    :param default_geometry: Default geometry of the screen
    :param default_min_dur: Default minimal duration between frames in
    milliseconds
    :param default_max_dur: Default maximal duration between frames in
    milliseconds
    :param default_max_dur: Default maximal duration between frames in
    milliseconds
    :param default_cmd: Default program (with argument list) recorded
    :param default_loop_delay: Duration of the pause between two consecutive
    loops of the animation in milliseconds
    :return: Tuple made of the subcommand called (None, 'render' or 'record')
    and all parsed
    arguments
    """
    command_parser = argparse.ArgumentParser(add_help=False)
    command_parser.add_argument(
        '-c', '--command',
        help=(('specify the program to record with optional arguments '
               '(default: {})').format(default_cmd)),
        default=default_cmd,
        metavar='COMMAND',
    )

    still_frames_parser = argparse.ArgumentParser(add_help=False)
    still_frames_parser.add_argument(
        '-s', '--still-frames',
        help='output still frames instead of an animation. ',
        action='store_true'
    )

    template_parser = argparse.ArgumentParser(add_help=False)
    template_parser.add_argument(
        '-t', '--template',
        help=('set the SVG template used for rendering the SVG animation. '
              'TEMPLATE may either be one of the default templates ({}) '
              'or a path to a valid template.').format(', '.join(templates)),
        type=lambda name: termtosvg.anim.validate_template(name, templates),
        default=default_template,
        metavar='TEMPLATE'
    )
    geometry_parser = argparse.ArgumentParser(add_help=False)
    geometry_parser.add_argument(
        '-g', '--screen-geometry',
        help='geometry of the terminal screen used for rendering the animation.'
             ' The geometry must be given as the number of columns and the '
             'number of rows on the screen separated by the character "x". '
             'For example "82x19" for an 82 columns by 19 rows screen.',
        metavar='GEOMETRY',
        default=default_geometry,
        type=termtosvg.config.validate_geometry
    )
    min_duration_parser = argparse.ArgumentParser(add_help=False)
    min_duration_parser.add_argument(
        '-m', '--min-frame-duration',
        type=integral_duration_validation,
        metavar='MIN_DURATION',
        default=default_min_dur,
        help=('minimum duration of a frame in milliseconds (default: {}ms)'
              .format(default_min_dur))
    )

    if default_max_dur:
        default_max_dur_label = '{}ms'.format(default_max_dur)
    else:
        default_max_dur_label = 'No maximum value'

    max_duration_parser = argparse.ArgumentParser(add_help=False)
    max_duration_parser.add_argument(
        '-M', '--max-frame-duration',
        type=integral_duration_validation,
        metavar='MAX_DURATION',
        default=default_max_dur,
        help=('maximum duration of a frame in milliseconds (default: {})'
              .format(default_max_dur_label))
    )

    loop_delay_parser = argparse.ArgumentParser(add_help=False)
    loop_delay_parser.add_argument(
        '-D', '--loop-delay',
        type=integral_duration_validation,
        metavar='DELAY',
        default=default_loop_delay,
        help=(('duration in milliseconds of the pause between two consecutive '
               'loops of the animation (default: {}ms)')
              .format(default_loop_delay))
    )

    parser = argparse.ArgumentParser(
        prog='termtosvg',
        parents=[command_parser, loop_delay_parser, geometry_parser, min_duration_parser,
                 max_duration_parser, still_frames_parser, template_parser],
        usage=USAGE,
        epilog=EPILOG
    )
    parser.add_argument(
        'output_path',
        nargs='?',
        help='optional filename of the SVG animation. If --still-frame is '
             'specified, output_path should be the path of the directory where '
             'still frames will be stored. If missing, a random path '
             'will be automatically generated.',
        metavar='output_path'
    )
    if args:
        if args[0] == 'record':
            parser = argparse.ArgumentParser(
                description='record the session to a file in asciicast v2 format',
                parents=[command_parser, geometry_parser],
                usage=RECORD_USAGE
            )
            parser.add_argument(
                'output_path',
                nargs='?',
                help='optional filename of the cast file. If missing, a random '
                     'path will be automatically generated.',
                metavar='output_path'
            )
            return args[0], parser.parse_args(args[1:])

        if args[0] == 'render':
            parser = argparse.ArgumentParser(
                description='render an asciicast recording as an SVG animation',
                parents=[loop_delay_parser,  min_duration_parser,
                         max_duration_parser, still_frames_parser, template_parser],
                usage=RENDER_USAGE
            )
            parser.add_argument(
                'input_file',
                help='recording of a terminal session in asciicast v1 or v2 format'
            )
            parser.add_argument(
                'output_path',
                nargs='?',
                help='optional filename of the SVG animation. If --still-frame '
                     'is specified, output_path should be the path of the '
                     'directory where still frames will be stored. If '
                     'missing, a random path will be automatically generated.',
                metavar='output_path'
            )
            return args[0], parser.parse_args(args[1:])

    return None, parser.parse_args(args)


def record_subcommand(process_args, geometry, input_fileno, output_fileno,
                      cast_filename):
    """Save a terminal session as an asciicast recording"""
    from termtosvg.term import get_terminal_size, TerminalMode, record
    logger.info('Recording started, enter "exit" command or Control-D to end')
    if geometry is None:
        columns, lines = get_terminal_size(output_fileno)
    else:
        columns, lines = geometry
    with TerminalMode(input_fileno):
        # Do not write anything to stdout (print, logger...) while in this
        # context manager if the output of the process is set to stdout. We
        # do not want two processes writing to the same terminal.
        records = record(process_args, columns, lines, input_fileno,
                         output_fileno)
        with open(cast_filename, 'w') as cast_file:
            for record_ in records:
                print(record_.to_json_line(), file=cast_file)
    logger.info('Recording ended, cast file is {}'.format(cast_filename))


def render_subcommand(still, template, cast_filename, output_path,
                      min_frame_duration, max_frame_duration, loop_delay):
    """Render the animation from an asciicast recording"""
    from termtosvg.asciicast import read_records
    from termtosvg.term import timed_frames

    logger.info('Rendering started')
    asciicast_records = read_records(cast_filename)
    geometry, frames = timed_frames(asciicast_records, min_frame_duration,
                                    max_frame_duration, loop_delay)
    if still:
        termtosvg.anim.render_still_frames(frames=frames,
                                           geometry=geometry,
                                           directory=output_path,
                                           template=template)
        logger.info('Rendering ended, SVG frames are located at {}'
                    .format(output_path))
    else:
        termtosvg.anim.render_animation(frames=frames,
                                        geometry=geometry,
                                        filename=output_path,
                                        template=template)
        logger.info('Rendering ended, SVG animation is {}'.format(output_path))


def record_render_subcommand(process_args, still, template, geometry,
                             input_fileno, output_fileno, output_path,
                             min_frame_duration, max_frame_duration,
                             loop_delay):
    """Record and render the animation on the fly"""
    from termtosvg.term import get_terminal_size, TerminalMode, record, timed_frames

    logger.info('Recording started, enter "exit" command or Control-D to end')
    if geometry is None:
        columns, lines = get_terminal_size(output_fileno)
    else:
        columns, lines = geometry
    with TerminalMode(input_fileno):
        # Do not write anything to stdout (print, logger...) while in this
        # context manager if the output of the process is set to stdout. We
        # do not want two processes writing to the same terminal.
        asciicast_records = record(process_args, columns, lines, input_fileno,
                                   output_fileno)
        geometry, frames = timed_frames(asciicast_records, min_frame_duration,
                                        max_frame_duration, loop_delay)

        if still:
            termtosvg.anim.render_still_frames(frames, geometry, output_path,
                                               template)
            end_msg = 'Rendering ended, SVG frames are located at {}'
        else:
            termtosvg.anim.render_animation(frames, geometry, output_path,
                                            template)
            end_msg = 'Rendering ended, SVG animation is {}'

    logger.info(end_msg.format(output_path))


def main(args=None, input_fileno=None, output_fileno=None):
    if args is None:
        args = sys.argv
    if input_fileno is None:
        input_fileno = sys.stdin.fileno()
    if output_fileno is None:
        output_fileno = sys.stdout.fileno()

    console_handler = logging.StreamHandler(sys.stderr)
    console_handler.setLevel(logging.INFO)
    console_formatter = logging.Formatter('%(message)s')
    console_handler.setFormatter(console_formatter)
    logger.handlers = [console_handler]
    logger.setLevel(logging.INFO)

    templates = termtosvg.config.default_templates()
    default_template = 'powershell'
    default_cmd = os.environ.get('SHELL', 'sh')
    command, args = parse(args[1:], templates, default_template, None, 1,
                          None, default_cmd, DEFAULT_LOOP_DELAY)

    if command == 'record':
        if args.output_path is None:
            _, cast_filename = tempfile.mkstemp(prefix='termtosvg_',
                                                suffix='.cast')
        else:
            cast_filename = args.output_path
        process_args = shlex.split(args.command)
        record_subcommand(process_args, args.screen_geometry, input_fileno,
                          output_fileno, cast_filename)
    elif command == 'render':
        if args.output_path is None:
            if args.still_frames:
                output_path = tempfile.mkdtemp(prefix='termtosvg_')
            else:
                _, output_path = tempfile.mkstemp(prefix='termtosvg_',
                                                  suffix='.svg')
        else:
            output_path = args.output_path
            if args.still_frames:
                try:
                    os.mkdir(output_path)
                except FileExistsError:
                    if not os.path.isdir(output_path):
                        raise

        render_subcommand(args.still_frames, args.template, args.input_file,
                          output_path, args.min_frame_duration,
                          args.max_frame_duration, args.loop_delay)
    else:
        if args.output_path is None:
            if args.still_frames:
                output_path = tempfile.mkdtemp(prefix='termtosvg_')
            else:
                _, output_path = tempfile.mkstemp(prefix='termtosvg_',
                                                  suffix='.svg')
        else:
            output_path = args.output_path
            if args.still_frames:
                try:
                    os.mkdir(output_path)
                except FileExistsError:
                    if not os.path.isdir(output_path):
                        raise

        process_args = shlex.split(args.command)
        record_render_subcommand(process_args, args.still_frames, args.template,
                                 args.screen_geometry, input_fileno,
                                 output_fileno, output_path,
                                 args.min_frame_duration,
                                 args.max_frame_duration,
                                 args.loop_delay)

    for handler in logger.handlers:
        handler.close()



import unittest

from termtosvg.asciicast import AsciiCastV2Header, AsciiCastV2Event, AsciiCastV2Record, \
                                AsciiCastV2Theme, AsciiCastError, _read_v1_records


class TestAsciicast(unittest.TestCase):
    def test_AsciiCastV2Theme(self):
        def palette(n):
            return ':'.join(['#000000'] * n)

        failure_test_cases = [
            ('invalid foreground color', None, '#ABCDEF', palette(8)),
            ('invalid hex number', '#BCDEFG', '#ABCDEF', palette(8)),
            ('invalid foreground color', '#123456', None, palette(16)),
            ('invalid palette', '#123456', '#XXXXXX', None),
            ('incomplete palette', '#123456', '#ABCDEF', palette(7)),
        ]
        for case, fg, bg, colors in failure_test_cases:
            with self.subTest(case=case):
                with self.assertRaises(AsciiCastError):
                    AsciiCastV2Theme(fg, bg, colors)

    cast_v2_lines = [
        # Header: Missing theme
        '{"version": 2, "width": 212, "height": 53}',
        # Header: 8 color theme
        '{"version": 2, "width": 212, "height": 53, "theme": {"fg": "#000000", "bg": "#AAAAAA", "palette": "#000000:#111111:#222222:#333333:#444444:#555555:#666666:#777777"}}',
        # Header: 16 color theme
        '{"version": 2, "width": 212, "height": 53, "theme": {"fg": "#000000", "bg": "#AAAAAA", "palette": "#000000:#111111:#222222:#333333:#444444:#555555:#666666:#777777:#888888:#999999:#AAAAAA:#bbbbbb:#CCCCCC:#DDDDDD:#EEEEEE:#ffffff"}}',
        # Header: additional values
        '{"version": 2, "width": 212, "height": 53, "timestamp": 123456798123}',
        # Header: idle_time_limit
        '{"version": 2, "width": 212, "height": 53, "timestamp": 123456798123, "idle_time_limit": 42}',
        # Float idle time limit (https://github.com/nbedos/termtosvg/issues/97)
        '{"version": 2, "width": 212, "height": 53, "idle_time_limit": 1.234}',
        # Event: Non printable characters
        '[0.010303, "o", "\\u001b[1;31mnico \\u001b[0;34m~\\u001b[0m"]',
        # Event: Unicode characters
        '[1.146397, "o", "❤ ☀ ☆ ☂ ☻ ♞ ☯ ☭ ☢ € →"]',
        # Event: time is an integer and not a float
        '[2, "o", "\\r\\n"]',
    ]

    color_theme_8 = AsciiCastV2Theme(fg='#000000', bg='#AAAAAA', palette='#000000:#111111:'
                                     '#222222:#333333:#444444:#555555:#666666:#777777')
    color_theme_16 = AsciiCastV2Theme(fg='#000000', bg='#AAAAAA', palette='#000000:#111111:'
                                      '#222222:#333333:#444444:#555555:#666666:#777777:#888888:'
                                      '#999999:#AAAAAA:#bbbbbb:#CCCCCC:#DDDDDD:#EEEEEE:#ffffff')
    cast_v2_events = [
        AsciiCastV2Header(2, 212, 53, None),
        AsciiCastV2Header(2, 212, 53, color_theme_8),
        AsciiCastV2Header(2, 212, 53, color_theme_16),
        AsciiCastV2Header(2, 212, 53, None),
        AsciiCastV2Header(2, 212, 53, None, 42),
        AsciiCastV2Header(2, 212, 53, None, 1.234),
        AsciiCastV2Event(0.010303, 'o', '\u001b[1;31mnico \u001b[0;34m~\u001b[0m', None),
        AsciiCastV2Event(1.146397, 'o', '❤ ☀ ☆ ☂ ☻ ♞ ☯ ☭ ☢ € →', None),
        AsciiCastV2Event(2, 'o', '\r\n', None),
    ]

    def test_from_json(self):
        test_cases = zip(TestAsciicast.cast_v2_lines, TestAsciicast.cast_v2_events)
        for index, (line, event) in enumerate(test_cases):
            with self.subTest(case='line #{}'.format(index)):
                self.assertEqual(event, AsciiCastV2Record.from_json_line(line))

        failure_test_cases = [
            ('header v2: invalid version', '{"version": "x", "width": 212, "height": 53}'),
            ('header v2: invalid width', '{"version": 2, "width": "x", "height": 53}'),
            ('header v2: invalid height', '{"version": 2, "width": "212", "height": null}'),
            ('header v2: wrong number of attributes', '{"version": 2}'),
            ('event v2: invalid time', '["x", "o", "ls"]'),
            ('event v2: invalid event_type', '[2.0, 123, "ls"]'),
            ('event v2: invalid event_data', '[2.0, 123, 42]'),
            ('event v2: invalid number of attributes', '[2.0, 123, "ls", "ls"]'),
            ('invalid record', '{"aaa": 42}'),
        ]
        for case, line in failure_test_cases:
            with self.subTest(case=case):
                with self.assertRaises(AsciiCastError):
                    AsciiCastV2Record.from_json_line(line)

    def test_to_json(self):
        test_cases = zip(TestAsciicast.cast_v2_lines, TestAsciicast.cast_v2_events)
        for index, (line, event) in enumerate(test_cases):
            # extra value test case only useful for from_json
            if 'timestamp' in line:
                continue
            with self.subTest(case='record #{}'.format(index)):
                self.assertEqual(event.to_json_line(), line)

    cast_v1_lines = '\r\n'.join(['{',
                                 '  "version": 1,',
                                 '  "width": 212,',
                                 '  "height": 53,',
                                 '  "duration": 2,',
                                 '  "command": "/bin/bash",',
                                 '  "title": "",',
                                 '  "env": {},',
                                 '  "stdout": [',
                                 '    [0.010303, "\\u001b[1;31mnico \\u001b[0;34m~\\u001b[0m"],',
                                 '    [1.136094, "❤ ☀ ☆ ☂ ☻ ♞ ☯ ☭ ☢ € →"],',
                                 '    [0.853603, "\\r\\n"]',
                                 '  ]',
                                 '}'])

    cast_v1_events = [
        AsciiCastV2Header(2, 212, 53, None),
        AsciiCastV2Event(0.010303, 'o', '\u001b[1;31mnico \u001b[0;34m~\u001b[0m',
                         None),
        AsciiCastV2Event(1.146397, 'o', '❤ ☀ ☆ ☂ ☻ ♞ ☯ ☭ ☢ € →', None),
        AsciiCastV2Event(2, 'o', '\r\n', None),
    ]

    def test__read_v1_records(self):
        test_cases = zip(TestAsciicast.cast_v1_lines,
                         TestAsciicast.cast_v1_events,
                         _read_v1_records(TestAsciicast.cast_v1_lines))
        for index, (line, expected_event, event) in enumerate(test_cases):
            with self.subTest(case='record #{}'.format(index)):
                if isinstance(event, AsciiCastV2Header):
                    self.assertEqual(event, expected_event)
                elif isinstance(event, AsciiCastV2Event):
                    self.assertAlmostEqual(event.time, expected_event.time, delta=10e-10)
                    self.assertEqual(event.event_data, expected_event.event_data)
                    self.assertEqual(event.event_type, expected_event.event_type)
                    self.assertEqual(event.duration, expected_event.duration)

        failure_test_cases = [
            ('invalid version', '{"version": null, "width": 212, "height": 53, "duration": 2, "stdout": []}'),
            ('invalid width', '{"version": 1, "width": null, "height": 53, "duration": 2, "stdout": []}'),
            ('invalid height', '{"version": 1, "width": 212, "height": "xx", "duration": 2, "stdout": []}'),
            ('invalid stdout', '{"version": 1, "width": 212, "height": 53, "duration": 2, "stdout": 42}'),
            ('missing version', '{"width": 212, "height": 53, "duration": 2, "stdout": []}'),
            ('missing height', '{"version": 1, "width": 212, "duration": 2, "stdout": []}'),
            ('invalid event', '{"version": 1, "width": 212, "height": 53, "duration": 2, "stdout": [[]]}'),
            ('invalid event', '{"version": 1, "width": 212, "height": 53, "duration": 2, "stdout": [[0]]}'),
            ('invalid event', '{"version": 1, "width": 212, "height": 53, "duration": 2, "stdout": [[0, "a", 2]]}'),
            ('invalid event duration', '{"version": 1, "width": 212, "height": 53, "duration": 2, "stdout": [["aaa", "a"]]}'),
            ('invalid JSON (header)','#####'),
            ('invalid JSON (event)', '{"version": 1, "width": 212, "height": 53, "duration": 2, "stdout": [[###]}'),
        ]

        for case, data in failure_test_cases:
            with self.subTest(case=case):
                with self.assertRaises(AsciiCastError):
                    for _ in _read_v1_records(data):
                        pass



"""UNIX terminal recording functionalities

This module exposes functions for
    - recording the output of a shell process in asciicast v2 format (`record`)
    - producing frames (2D array of CharacterCell) from the raw output of
    `record` (`timed_frames`)

A context manager named `TerminalMode` is also provided and is to be used with
the `record` function to ensure that the terminal state is always properly
restored, otherwise a failure during a call to `record` could render the
terminal unusable.
"""

import codecs
import datetime
import fcntl
import os
import pty
import select
import struct
import termios
import tty
from collections import defaultdict, namedtuple
from typing import Iterator

import pyte
import pyte.screens

from termtosvg import anim
from termtosvg.asciicast import AsciiCastV2Event, AsciiCastV2Header

TimedFrame = namedtuple('TimedFrame', ['time', 'duration', 'buffer'])


class TerminalMode:
    """Save terminal mode and size on entry, restore them on exit

    This context manager exists to ensure that the state of the terminal is
    properly restored when functions like `_record` (which relies on setting
    the terminal mode to raw and changing the geometry of the screen) fail.
    """
    def __init__(self, fileno):
        self.fileno = fileno
        self.mode = None
        self.ttysize = None

    def __enter__(self):
        try:
            self.mode = tty.tcgetattr(self.fileno)
        except tty.error:
            pass

        try:
            columns, lines = os.get_terminal_size(self.fileno)
        except OSError:
            pass
        else:
            self.ttysize = struct.pack("HHHH", lines, columns, 0, 0)

        return self.mode, self.ttysize

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.ttysize is not None:
            fcntl.ioctl(self.fileno, termios.TIOCSWINSZ, self.ttysize)

        if self.mode is not None:
            tty.tcsetattr(self.fileno, tty.TCSAFLUSH, self.mode)


def _record(process_args, columns, lines, input_fileno, output_fileno):
    """Record raw input and output of a process

    This function forks the current process. The child process runs the command
    specified by 'process_args' which is a session leader and has a controlling
    terminal and is run in the background. The parent process, which runs in
    the foreground, transmits data between the standard input, output and the
    child process and logs it. From the user point of view, it appears they are
    communicating with the process they intend to record (through their
    terminal emulator) when in fact they communicate with our parent process
    which logs all data exchanges with the user

    The implementation of this method is mostly copied from the pty.spawn
    function of the CPython standard library. It has been modified in order to
    make the record function a generator.
    See https://github.com/python/cpython/blob/master/Lib/pty.py

    :param process_args: List of arguments to run the process to be recorded
    :param columns: Initial number of columns of the terminal
    :param lines: Initial number of lines of the terminal
    :param input_fileno: File descriptor of the input data stream
    :param output_fileno: File descriptor of the output data stream
    """
    pid, master_fd = pty.fork()
    if pid == 0:
        # Child process - this call never returns
        os.execlp(process_args[0], *process_args)

    # Parent process
    # Set the terminal size for master_fd
    ttysize = struct.pack("HHHH", lines, columns, 0, 0)
    fcntl.ioctl(master_fd, termios.TIOCSWINSZ, ttysize)

    try:
        tty.setraw(input_fileno)
    except tty.error:
        pass

    for data, time in _capture_output(input_fileno, output_fileno, master_fd):
        yield data, time

    os.close(master_fd)

    _, child_exit_status = os.waitpid(pid, 0)
    return child_exit_status


def _capture_output(input_fileno, output_fileno, master_fd, buffer_size=1024):
    """Send data from input_fileno to master_fd and send data from master_fd to
    output_fileno and to the caller

    The implementation of this method is mostly copied from the pty.spawn
    function of the CPython standard library. It has been modified in order to
    make the `record` function a generator.

    See https://github.com/python/cpython/blob/master/Lib/pty.py
    """
    rlist = [input_fileno, master_fd]
    xlist = [input_fileno, output_fileno, master_fd]

    xfds = []
    while not xfds:
        rfds, _, xfds = select.select(rlist, [], xlist)
        for fd in rfds:
            try:
                data = os.read(fd, buffer_size)
            except OSError:
                xfds.append(fd)
                continue

            if not data:
                xfds.append(fd)
                continue

            if fd == input_fileno:
                write_fileno = master_fd
            else:
                write_fileno = output_fileno
                yield data, datetime.datetime.now()

            while data:
                n = os.write(write_fileno, data)
                data = data[n:]


def _group_by_time(event_records, min_rec_duration, max_rec_duration, last_rec_duration):
    """Merge event records together if they are close enough

    The time elapsed between two consecutive event records returned by this
    function is guaranteed to be at least min_rec_duration.

    The duration of each record is also computed. Any record with a duration
    greater than `max_rec_duration` will see its duration reduce to this value.
    The duration of the last record can't be computed and is simply set to
    `last_rec_duration`.

    :param event_records: Sequence of records in asciicast v2 format
    :param min_rec_duration: Minimum time between two records returned by the
    function in milliseconds.
    :param max_rec_duration: Maximum duration of a record in milliseconds
    :param last_rec_duration: Duration of the last record in milliseconds
    :return: Sequence of records with duration
    """
    # TODO: itertools.accumulate?
    current_string = ''
    current_time = 0
    dropped_time = 0

    if max_rec_duration:
        max_rec_duration /= 1000

    for event_record in event_records:
        assert isinstance(event_record, AsciiCastV2Event)
        # Silently ignoring the duration on input records is a source
        # of confusion so fail hard if the duration is set
        assert event_record.duration is None
        if event_record.event_type != 'o':
            continue

        time_between_events = event_record.time - (current_time + dropped_time)
        if time_between_events * 1000 >= min_rec_duration:
            if max_rec_duration:
                if max_rec_duration < time_between_events:
                    dropped_time += time_between_events - max_rec_duration
                    time_between_events = max_rec_duration
            accumulator_event = AsciiCastV2Event(time=current_time,
                                                 event_type='o',
                                                 event_data=current_string,
                                                 duration=time_between_events)
            yield accumulator_event
            current_string = ''
            current_time += time_between_events

        current_string += event_record.event_data

    accumulator_event = AsciiCastV2Event(time=current_time,
                                         event_type='o',
                                         event_data=current_string,
                                         duration=last_rec_duration / 1000)
    yield accumulator_event


def record(process_args, columns, lines, input_fileno, output_fileno):
    """Record a process in asciicast v2 format

    The records returned by this method are:
        - a single header containing configuration information
        - multiple event records made of data captured from the terminal and
        timing information (except for record duration which needs to be
        computed separately)

    :param process_args: Arguments required to spawn the process (list of
    string)
    :param columns: Width of the terminal screen (integer)
    :param lines: Height of the terminal screen (integer)
    :param input_fileno: File descriptor that will be used as the standard
    input of the process
    :param output_fileno: File descriptor that will be used as the standard
    output of the process

    When using `sys.stdout.fileno()` for `output_fileno` there is a risk
    that the terminal is left in an unusable state if `record` fails. To
    prevent this, `record` should be called inside the `TerminalMode`
    context manager.
    """
    yield AsciiCastV2Header(version=2, width=columns, height=lines, theme=None)

    # TODO: why start != 0?
    start = None
    utf8_decoder = codecs.getincrementaldecoder('utf-8')('replace')
    for data, time in _record(process_args, columns, lines, input_fileno, output_fileno):
        if start is None:
            start = time

        yield AsciiCastV2Event(time=(time - start).total_seconds(),
                               event_type='o',
                               event_data=utf8_decoder.decode(data),
                               duration=None)


def timed_frames(records, min_frame_dur=1, max_frame_dur=None, last_frame_dur=1000):
    """Return a tuple made of the geometry of the screen and a generator of
    instances of TimedFrame computed from asciicast records

    Asciicast records are first coalesced so that the mininum duration between
    two frames is at least `min_frame_dur` milliseconds. Events with a duration
    greater than `max_frame_dur` will see their duration reduced to that value.

    The duration of all frames lasting until the end of the animation
    will be adjusted so that the last frame of the animation lasts
    `last_frame_dur`

    :param records: Terminal session record in Asciicast v2 format
    :param min_frame_dur: Minimum frame duration in milliseconds (integer)
    :param min_frame_dur: Minimum frame duration in milliseconds (integer)
    :param max_frame_dur: Maximum frame duration in milliseconds (None or
    integer)
    :param last_frame_dur: Duration of the last frame of the animation
    (integer)
    """
    if not isinstance(records, Iterator):
        records = iter(records)

    header = next(records)
    assert isinstance(header, AsciiCastV2Header)

    if not max_frame_dur and header.idle_time_limit:
        max_frame_dur = int(header.idle_time_limit * 1000)

    def generator():
        screen = pyte.Screen(header.width, header.height)
        stream = pyte.Stream(screen)
        timed_records = _group_by_time(records, min_frame_dur, max_frame_dur,
                                       last_frame_dur)

        for record_ in timed_records:
            assert isinstance(record_, AsciiCastV2Event)
            for char in record_.event_data:
                stream.feed(char)
            yield TimedFrame(int(1000 * record_.time),
                             int(1000 * record_.duration),
                             _screen_buffer(screen))

    return (header.width, header.height), generator()


def _screen_buffer(screen):
    assert isinstance(screen, pyte.Screen)

    buffer = defaultdict(dict)
    for row in range(screen.lines):
        buffer[row] = {
            column: anim.CharacterCell.from_pyte(screen.buffer[row][column])
            for column in screen.buffer[row]
        }

    if not screen.cursor.hidden:
        row, column = screen.cursor.y, screen.cursor.x
        try:
            data = screen.buffer[row][column].data
        except KeyError:
            data = ' '

        cursor_char = pyte.screens.Char(data=data,
                                        fg=screen.cursor.attrs.fg,
                                        bg=screen.cursor.attrs.bg,
                                        reverse=True)
        buffer[row][column] = anim.CharacterCell.from_pyte(cursor_char)
    return buffer


def get_terminal_size(fileno):
    try:
        columns, lines = os.get_terminal_size(fileno)
    except OSError:
        columns, lines = 80, 24

    return columns, lines



import unittest

from termtosvg.tests.test_anim import TestAnim
from termtosvg.tests.test_asciicast import TestAsciicast
from termtosvg.tests.test_config import TestConf
from termtosvg.tests.test_main import TestMain
from termtosvg.tests.test_term import TestTerm



import itertools
import os
import time
import unittest
from unittest.mock import MagicMock, patch

import pyte

import termtosvg.anim as anim
from termtosvg import term
from termtosvg.asciicast import AsciiCastV2Event, AsciiCastV2Header, AsciiCastV2Theme

commands = [
    'echo $SHELL && sleep 0.1;\r\n',
    'date && sleep 0.1;\r\n',
    'uname && sleep 0.1;\r\n',
    'w',
    'h',
    'o',
    'a',
    'm',
    'i\r\n',
    'exit;\r\n'
]


THEME = AsciiCastV2Theme('#000000', '#FFFFFF', ':'.join(['#123456'] * 16))
CURSOR_CHAR = anim.CharacterCell(' ', 'background', 'foreground')

unittest.TestCase.maxDiff = None


class TestTerm(unittest.TestCase):
    def test__record(self):
        fd_in_read, fd_in_write = os.pipe()
        fd_out_read, fd_out_write = os.pipe()

        lines = 24
        columns = 80

        pid = os.fork()
        if pid == 0:
            # Child process
            for line in commands:
                os.write(fd_in_write, line.encode('utf-8'))
                time.sleep(0.060)
            os._exit(0)

        # Parent process
        with term.TerminalMode(fd_in_read):
            for _ in term._record(['sh'], columns, lines, fd_in_read, fd_out_write):
                pass

        os.waitpid(pid, 0)
        for fd in fd_in_read, fd_in_write, fd_out_read, fd_out_write:
            os.close(fd)

    def test_record(self):
        # Use pipes in lieu of stdin and stdout
        fd_in_read, fd_in_write = os.pipe()
        fd_out_read, fd_out_write = os.pipe()

        lines = 24
        columns = 80

        pid = os.fork()
        if pid == 0:
            # Child process
            for line in commands:
                os.write(fd_in_write, line.encode('utf-8'))
                time.sleep(0.060)
            os._exit(0)

        # Parent process
        with term.TerminalMode(fd_in_read):
            for _ in term.record(['sh'], columns, lines, fd_in_read, fd_out_write):
                pass

        os.waitpid(pid, 0)
        for fd in fd_in_read, fd_in_write, fd_out_read, fd_out_write:
            os.close(fd)

    def test__buffer_simple_events(self):
        escape_sequences = ['{}\r\n'.format(i) for i in range(5)]

        screen = pyte.Screen(80, 24)
        stream = pyte.Stream(screen)
        for count, escape_sequence in enumerate(escape_sequences):
            with self.subTest(case='Simple events (record #{})'.format(count)):
                stream.feed(escape_sequence)
                buffer = term._screen_buffer(screen)
                expected_buffer = {}
                for i in range(screen.lines):
                    if i <= count:
                        expected_buffer[i] = {0: anim.CharacterCell(str(i))}
                    elif i == count+1:
                        expected_buffer[i] = {0: CURSOR_CHAR}
                    else:
                        expected_buffer[i] = {}

                self.assertEqual(expected_buffer, buffer)

    def test__buffer_hidden_cursor(self):
        """Ensure hidden cursor don't appear in the buffer return by _redraw_buffer
        Ensure that only one cursor is present in the buffer"""
        #   '\u001b[?25h' : display cursor
        #   '\u001b[?25l' : hide cursor
        escape_sequences = [
            '\u001b[?25ha',
            '\r\n\u001b[?25lb',
            '\r\n\u001b[?25hc',
        ]

        screen = pyte.Screen(80, 24)
        stream = pyte.Stream(screen)

        expected_cursors = [
            ((1, 0), True),
            ((1, 1), False),
            ((1, 2), True),
        ]
        z = itertools.zip_longest(expected_cursors, escape_sequences)
        for count, ((cursor_pos, cursor_visible), escape_sequence) in enumerate(z):
            with self.subTest(case='Hidden cursor - item #{}'.format(count)):
                stream.feed(escape_sequence)
                buffer = term._screen_buffer(screen)
                column, line = cursor_pos
                if cursor_visible:
                    self.assertEqual(buffer[line][column], CURSOR_CHAR)
                # Ensure old cursors are deleted
                for row in buffer:
                    for column in buffer[row]:
                        if buffer[row][column].text == ' ':
                            self.assertEqual((column, row), cursor_pos)

    def test_timed_frames_simple_events(self):
        records = [AsciiCastV2Header(version=2, width=80, height=24, theme=THEME)] + \
                  [AsciiCastV2Event(time=i,
                                    event_type='o',
                                    event_data='{}\r\n'.format(i),
                                    duration=None)
                   for i in range(0, 2)]
        geometry, frames = term.timed_frames(records, 1, None, 42)

        self.assertEqual(geometry, (80, 24))

        expected_frames = [
            term.TimedFrame(0, 1000, {
                0: {0: anim.CharacterCell('0')},
                1: {0: CURSOR_CHAR},
            }),
            term.TimedFrame(1000, 42, {
                0: {0: anim.CharacterCell('0')},
                1: {0: anim.CharacterCell('1')},
                2: {0: CURSOR_CHAR},
            })
        ]

        z = itertools.zip_longest(expected_frames, frames)
        for (expected_frame, frame) in z:
            self.assertEqual(expected_frame.time, frame.time)
            self.assertEqual(expected_frame.duration, frame.duration)
            for row in frame.buffer:
                if row in expected_frame.buffer:
                    self.assertEqual(expected_frame.buffer[row],
                                     frame.buffer[row])
                else:
                    self.assertEqual({}, frame.buffer[row])

    def test_timed_frames_unprintable_chars(self):
        # Ensure zero width characters in terminal output does not result
        # in Pyte dropping all following data
        # Issue https://github.com/nbedos/termtosvg/issues/89

        # test_text = "e🕵️‍a"
        test_text = (b'e' +
                     b'\xf0\x9f\x95\xb5' +  # sleuth emoji
                     b'\xef\xb8\x8f' +      # variation selector 16
                     b'\xe2\x80\x8d' +      # zero width joiner
                     b'a').decode('utf-8')  # character that should be preserved

        records = [
            AsciiCastV2Header(version=2, width=80, height=24, theme=THEME),
            AsciiCastV2Event(0, 'o', test_text, None),
        ]
        _, events = term.timed_frames(records, 1, None, last_frame_dur=1000)

        frame = next(events)
        characters = ''.join(frame.buffer[0][col].text
                             for col in frame.buffer[0])

        # Ensure data following sleuth emoji wasn't ignored
        # (rstrip() removes blank cursor character at end of line)
        self.assertEqual(characters.rstrip()[-1], test_text[-1])

    def test_get_terminal_size(self):
        with self.subTest(case='Successful get_terminal_size call'):
            term_size_mock = MagicMock(return_value=(42, 84))
            with patch('os.get_terminal_size', term_size_mock):
                cols, lines, = term.get_terminal_size(-1)
                self.assertEqual(cols, 42)
                self.assertEqual(lines, 84)

    def test__group_by_time(self):
        event_records = [
            AsciiCastV2Event(0, 'o', '1', None),
            AsciiCastV2Event(5, 'o', '2', None),
            AsciiCastV2Event(8, 'o', '3', None),
            AsciiCastV2Event(20, 'o', '4', None),
            AsciiCastV2Event(21, 'o', '5', None),
            AsciiCastV2Event(30, 'o', '6', None),
            AsciiCastV2Event(31, 'o', '7', None),
            AsciiCastV2Event(32, 'o', '8', None),
            AsciiCastV2Event(33, 'o', '9', None),
            AsciiCastV2Event(43, 'o', '10', None),
        ]

        with self.subTest(case='maximum record duration'):
            grouped_event_records_max = [
                AsciiCastV2Event(0, 'o', '1', 5),
                AsciiCastV2Event(5, 'o', '23', 6),
                AsciiCastV2Event(11, 'o', '45', 6),
                AsciiCastV2Event(17, 'o', '6789', 6),
                AsciiCastV2Event(23, 'o', '10', 1.234),
            ]
            result = list(term._group_by_time(event_records, 5000, 6000, 1234))
            self.assertEqual(grouped_event_records_max, result)

        with self.subTest(case='no maximum record duration'):
            grouped_event_records_no_max = [
                AsciiCastV2Event(0, 'o', '1', 5),
                AsciiCastV2Event(5, 'o', '23', 15),
                AsciiCastV2Event(20, 'o', '45', 10),
                AsciiCastV2Event(30, 'o', '6789', 13),
                AsciiCastV2Event(43, 'o', '10', 1.234),
            ]
            result = list(term._group_by_time(event_records, 5000, None, 1234))
            self.assertEqual(grouped_event_records_no_max, result)



import io
import itertools
import pkgutil
import tempfile
import unittest
from collections import namedtuple

import pyte.screens
from lxml import etree

from termtosvg import anim
from termtosvg import term


TEMPLATE = pkgutil.get_data('termtosvg', '/data/templates/gjm8.svg')


def line(i):
    chars = []
    for c in 'line{}'.format(i):
        chars.append(anim.CharacterCell(c, '#123456', '#789012',
                                        False, False, False, False))
    return dict(enumerate(chars))


class TestAnim(unittest.TestCase):
    def test_from_pyte(self):
        pyte_chars = [
            # Simple mapping
            pyte.screens.Char('A', 'red', 'blue'),
            # Reverse colors
            pyte.screens.Char('B', 'red', 'blue', reverse=True),
            # Bold for foreground -> bright colors
            pyte.screens.Char('C', 'red', 'blue', bold=True),
            # Bold and reverse
            pyte.screens.Char('D', 'red', 'blue', bold=True, reverse=True),
            # Defaults
            pyte.screens.Char('E', 'default', 'default'),
            # Hexadecimal
            pyte.screens.Char('F', '008700', 'ABCDEF'),
            # Bright and bold
            pyte.screens.Char('G', 'brightgreen', 'ABCDEF', bold=True),
            # Italics
            pyte.screens.Char('H', 'red', 'blue', italics=True),
            # Underscore
            pyte.screens.Char('I', 'red', 'blue', underscore=True),
            # Strikethrough
            pyte.screens.Char('J', 'red', 'blue', strikethrough=True),
        ]

        char_cells = [
            anim.CharacterCell('A', 'color1', 'color4'),
            anim.CharacterCell('B', 'color4', 'color1'),
            anim.CharacterCell('C', 'color9', 'color4', bold=True),
            anim.CharacterCell('D', 'color4', 'color9', bold=True),
            anim.CharacterCell('E', 'foreground', 'background'),
            anim.CharacterCell('F', '#008700', '#ABCDEF'),
            anim.CharacterCell('G', 'color10', '#ABCDEF', bold=True),
            anim.CharacterCell('H', 'color1', 'color4', italics=True),
            anim.CharacterCell('I', 'color1', 'color4', underscore=True),
            anim.CharacterCell('J', 'color1', 'color4', strikethrough=True),
        ]

        for pyte_char, cell_char in zip(pyte_chars, char_cells):
            with self.subTest(case=pyte_char):
                self.assertEqual(anim.CharacterCell.from_pyte(pyte_char), cell_char)

    def test__render_line_bg_colors_xml(self):
        cell_width = 8
        screen_line = {
            0: anim.CharacterCell('A', 'black', 'red'),
            1: anim.CharacterCell('A', 'black', 'red'),
            3: anim.CharacterCell('A', 'black', 'red'),
            4: anim.CharacterCell('A', 'black', 'blue'),
            6: anim.CharacterCell('A', 'black', 'blue'),
            7: anim.CharacterCell('A', 'black', 'blue'),
            8: anim.CharacterCell('A', 'black', 'green'),
            9: anim.CharacterCell('A', 'black', 'red'),
            10: anim.CharacterCell('A', 'black', 'red'),
            11: anim.CharacterCell('A', 'black', '#123456'),
        }

        rectangles = anim._render_line_bg_colors(screen_line=screen_line,
                                                 height=0,
                                                 cell_height=1,
                                                 cell_width=cell_width)

        def key(r):
            return r.attrib['x']

        rect_0, rect_3, rect_4, rect_6, rect_8, rect_9, rect_11 = sorted(rectangles, key=key)

        self.assertEqual(rect_0.attrib['x'], '0')
        self.assertEqual(rect_0.attrib['width'], '16')
        self.assertEqual(rect_0.attrib['height'], '1')
        self.assertEqual(rect_0.attrib['class'], 'red')
        self.assertEqual(rect_3.attrib['x'], '24')
        self.assertEqual(rect_3.attrib['width'], '8')
        self.assertEqual(rect_4.attrib['x'], '32')
        self.assertEqual(rect_6.attrib['x'], '48')
        self.assertEqual(rect_6.attrib['width'], '16')
        self.assertEqual(rect_6.attrib['class'], 'blue')
        self.assertEqual(rect_8.attrib['x'], '64')
        self.assertEqual(rect_8.attrib['class'], 'green')
        self.assertEqual(rect_9.attrib['x'], '72')
        self.assertEqual(rect_11.attrib['fill'], '#123456')

    def test__render_characters(self):
        screen_line = {
            0: anim.CharacterCell('A', 'red', 'white'),
            1: anim.CharacterCell('B', 'blue', 'white'),
            2: anim.CharacterCell('C', 'blue', 'white'),
            7: anim.CharacterCell('D', '#00FF00', 'white'),
            8: anim.CharacterCell('E', '#00FF00', 'white'),
            9: anim.CharacterCell('F', '#00FF00', 'white'),
            10: anim.CharacterCell('G', '#00FF00', 'white'),
            20: anim.CharacterCell('H', 'black', 'black', bold=True),
            30: anim.CharacterCell('I', 'black', 'black', italics=True),
            40: anim.CharacterCell('J', 'black', 'black', underscore=True),
            50: anim.CharacterCell('K', 'black', 'black', strikethrough=True),
            60: anim.CharacterCell('L', 'black', 'black', underscore=True, strikethrough=True),
        }

        with self.subTest(case='Content'):
            cell_width = 8
            texts = {t.text: t for t in anim._render_characters(screen_line, cell_width)}

            self.assertEqual(texts['A'].attrib['class'], 'red')
            self.assertEqual(texts['A'].attrib['x'], '0')
            # Style attributes should not appear for normal text
            self.assertNotIn('font-weight', texts['A'].attrib)
            self.assertNotIn('font-style', texts['A'].attrib)
            self.assertNotIn('text-decoration', texts['A'].attrib)
            self.assertEqual(texts['BC'].attrib['class'], 'blue')
            self.assertEqual(texts['BC'].attrib['x'], '8')
            self.assertEqual(texts['DEFG'].attrib['fill'], '#00FF00')
            self.assertEqual(texts['DEFG'].attrib['x'], '56')
            self.assertEqual(texts['H'].attrib['font-weight'], 'bold')
            self.assertEqual(texts['I'].attrib['font-style'], 'italic')
            self.assertIn('underline', texts['J'].attrib['text-decoration'].split())
            self.assertIn('line-through', texts['K'].attrib['text-decoration'].split())
            self.assertIn('underline', texts['L'].attrib['text-decoration'].split())
            self.assertIn('line-through', texts['L'].attrib['text-decoration'].split())

    def test_ConsecutiveWithSameAttributes(self):
        testClass = namedtuple('testClass', ['field1', 'field2'])
        test_cases = [
            (0, testClass('a', 'b')),
            (2, testClass('a', 'b')),
            (3, testClass('a', 'b')),
            (4, testClass('b', 'b')),
            (5, testClass('b', 'a')),
            (10, testClass('c', 'd')),
            (11, testClass('c', 'd')),
        ]

        expected_results = [
            (0, {'field1': 'a', 'field2': 'b'}),
            (2, {'field1': 'a', 'field2': 'b'}),
            (2, {'field1': 'a', 'field2': 'b'}),
            (4, {'field1': 'b', 'field2': 'b'}),
            (5, {'field1': 'b', 'field2': 'a'}),
            (10, {'field1': 'c', 'field2': 'd'}),
            (10, {'field1': 'c', 'field2': 'd'}),
        ]

        key = anim.ConsecutiveWithSameAttributes(['field1', 'field2'])
        for case, result in zip(test_cases, expected_results):
            with self.subTest(case=case):
                self.assertEqual(key(case), result)

    def test__render_timed_frame(self):
        frames = [
            term.TimedFrame(1, 1, {
                1: line(1),
            }),
            term.TimedFrame(2, 1, {
                1: line(1),
                2: line(2),
            }),
            term.TimedFrame(3, 1, {
                1: line(1),
                2: line(2),
                3: line(3),
            }),
            term.TimedFrame(4, 1, {
                1: line(1),
                2: line(2),
                3: line(3),
                4: line(4),
            }),
            # Definition reuse
            term.TimedFrame(5, 1, {
                1: line(1),
                2: line(2),
                3: line(3),
                5: line(4),
            }),
        ]

        all_definitions = {}
        for frame in frames:
            group, new_defs = anim._render_timed_frame(offset=0,
                                                       buffer=frame.buffer,
                                                       cell_width=8,
                                                       cell_height=17,
                                                       definitions={})
            all_definitions.update(new_defs)

        assert(len(all_definitions) == 4)

    def test_render_animation(self):
        frames = [
            term.TimedFrame(0, 60, {
                0: line(0),
            }),
            term.TimedFrame(60, 60, {
                0: line(0),
                1: line(1),
            }),
            term.TimedFrame(120, 60, {
                2: line(2),
                3: line(3),
            }),
        ]
        _, filename = tempfile.mkstemp(prefix='termtosvg_', suffix='.svg')
        anim.render_animation(frames, (80, 24), filename, TEMPLATE)
        with open(filename) as f:
            anim.validate_svg(f)

    def test__render_still_frames(self):
        def line(s):
            return dict(enumerate([anim.CharacterCell(c) for c in s]))

        frames = [
            term.TimedFrame(0, 0, {
                1: line('a'),
                2: line('b'),
            }),
            term.TimedFrame(120, 120, {
                2: line('b'),
                3: line('c'),
                4: line('d'),
            }),
            term.TimedFrame(240, 60, {
                2: line('b'),
                3: line('c'),
                4: line('d'),
                5: line('e'),
            }),
            term.TimedFrame(300, 60, {
                2: line('b'),
                3: line('c'),
                4: line('d'),
                5: line('f'),
            }),
        ]

        root = anim._render_preparation((80, 24), TEMPLATE, 9, 17)
        frame_generator = anim._render_still_frames(frames, root, 9, 17)

        def extract_text_content(frame):
            svg_screen_elem = frame.find('.//{{{namespace}}}svg[@id="screen"]'
                                         .format(namespace=anim.SVG_NS))
            return {elem.text for elem in svg_screen_elem.findall('.//text')}

        expected_texts_by_frame = [
            {'a', 'b'},
            {'b', 'c', 'd'},
            {'b', 'c', 'd', 'e'},
            {'b', 'c', 'd', 'f'},
        ]

        z = itertools.zip_longest(expected_texts_by_frame, frame_generator)
        for count, (texts, frame) in enumerate(z):
            with self.subTest(case='Frame #{}'.format(count)):
                anim.validate_svg(io.BytesIO(etree.tostring(frame)))
                self.assertEqual(texts, extract_text_content(frame))

    def test__embed_css(self):
        test_cases = [{'animation_duration': None, 'timings': {1: 100, 2: 200}},
                      {'animation_duration': 42, 'timings': {12: 100, 33: 200}} ]
        for args in test_cases:
            with self.subTest(case=args['animation_duration']):
                tree = etree.parse(io.BytesIO(TEMPLATE))
                root = tree.getroot()
                anim._embed_css(root, **args)
                assert b'{{' not in etree.tostring(root)

    def test_validate_svg(self):
        failure_test_cases = [
            '',
            '<svg>',
            '</svg>',
            '</svg></a>',
            None
        ]
        for case in failure_test_cases:
            with self.subTest(case=case):
                with self.assertRaises(ValueError):
                    anim.validate_svg(io.StringIO(case))

        success_test_cases = [
            TEMPLATE,
        ]
        for bytes_svg in success_test_cases:
            with io.BytesIO(bytes_svg) as bstream:
                anim.validate_svg(bstream)



import os
import shutil
import tempfile
import time
import unittest

import termtosvg.main
import termtosvg.config

SHELL_INPUT = [
    'echo $SHELL && sleep 0.1;\r\n',
    'date && sleep 0.1;\r\n',
    'uname && sleep 0.1;\r\n',
    'w',
    'h',
    'o',
    'a',
    'm',
    'i\r\n',
    'echo -e "\\033[1;31mbright red fg\\033[0m"\r\n',
    'echo -e "\\033[1;41mbright red bg\\033[0m"\r\n',
    'echo -e "\\033[1mbold\\033[0m"\r\n',
    'echo -e "\\033[3mitalics\\033[0m"\r\n',
    'echo -e "\\033[4munderscore\\033[0m"\r\n',
    'echo -e "\\033[9mstrikethrough\\033[0m"\r\n',
    'exit;\r\n'
]


class TestMain(unittest.TestCase):
    test_cases = [
        [],
        ['-c', 'sh'],
        ['--screen-geometry', '82x19'],
        ['-D', '1234'],
        ['--loop-delay', '1234'],
        ['-g', '82x19'],
        ['--template', 'plain'],
        ['-t', 'plain'],
        ['-s'],
        ['--still-frames'],
        ['--screen-geometry', '82x19', '--template', 'plain'],
        ['output_path', '-g', '82x19', '-t', 'plain', '-c', 'date', '-s'],
        ['--screen-geometry', '82x19', '--template', 'plain', '-s', 'output_path'],
        ['-g', '82x19', '-t', 'plain'],
        ['-m', '42', '-M', '100'],
        ['--min-frame-duration', '42ms', '--max-frame-duration', '100'],
        ['record'],
        ['record', '-c', 'ls'],
        ['record', 'output_path'],
        ['record', 'output_path', '--screen-geometry', '82x19'],
        ['record', '--screen-geometry', '82x19'],
        ['render', 'input_filename'],
        ['render', 'input_filename'],
        ['render', 'input_filename', '--template', 'plain'],
        ['render', 'input_filename', 'output_path'],
        ['render', 'input_filename', 'output_path', '--template', 'plain'],
        ['render', 'input_filename', 'output_path', '-t', 'plain', '-m', '42', '-M', '100'],
        ['render', 'input_filename', 'output_path', '-t', 'plain', '-m', '42', '-s', '-M', '100'],
    ]

    def test_parse(self):
        for args in self.test_cases:
            with self.subTest(case=args):
                cmd, parsed_args = termtosvg.main.parse(
                    args=args,
                    templates={'plain': b''},
                    default_template='plain',
                    default_geometry='48x95',
                    default_min_dur=2,
                    default_max_dur=None,
                    default_cmd='sh',
                    default_loop_delay=1000
                )

    @staticmethod
    def run_main(args, process_input):
        # Use pipes in lieu of stdin and stdout
        fd_in_read, fd_in_write = os.pipe()
        fd_out_read, fd_out_write = os.pipe()

        pid = os.fork()
        if pid == 0:
            # Child process
            for line in process_input:
                os.write(fd_in_write, line.encode('utf-8'))
                time.sleep(0.060)
            os._exit(0)

        termtosvg.main.main(args, fd_in_read, fd_out_write)

        os.waitpid(pid, 0)
        for fd in fd_in_read, fd_in_write, fd_out_read, fd_out_write:
            os.close(fd)

    def test_main(self):
        _, cast_filename = tempfile.mkstemp(prefix='termtosvg_', suffix='.cast')
        svg_filename = cast_filename[:-5] + '.svg'

        with self.subTest(case='record (no filename)'):
            args = ['termtosvg', 'record']
            TestMain.run_main(args, SHELL_INPUT)

        with self.subTest(case='record (with filename)'):
            args = ['termtosvg', 'record', cast_filename]
            TestMain.run_main(args, SHELL_INPUT)

        with self.subTest(case='record (with geometry)'):
            args = ['termtosvg', 'record', '--screen-geometry', '82x19']
            TestMain.run_main(args, SHELL_INPUT)

        with self.subTest(case='record (with command)'):
            args = ['termtosvg', 'record', '-c', 'date']
            TestMain.run_main(args, [])

        with self.subTest(case='render (no output filename)'):
            args = ['termtosvg', 'render', cast_filename]
            TestMain.run_main(args, [])

        with self.subTest(case='render (with output filename)'):
            args = ['termtosvg', 'render', cast_filename, svg_filename]
            TestMain.run_main(args, [])

        with self.subTest(case='render (with delay)'):
            args = ['termtosvg', 'render', cast_filename, '-D', '1234']
            TestMain.run_main(args, [])

        with self.subTest(case='render (with template)'):
            args = ['termtosvg', 'render', cast_filename, '--template', 'window_frame']
            TestMain.run_main(args, [])

        with self.subTest(case='render (still frames)'):
            args = ['termtosvg', 'render', cast_filename, '--still-frames']
            TestMain.run_main(args, [])

        with self.subTest(case='render (still frames with output directory)'):
            # Existing directory
            output_path = tempfile.mkdtemp(prefix='termtosvg')
            args = ['termtosvg', 'render', cast_filename, output_path, '-s']
            TestMain.run_main(args, [])

            # Non existing directory
            shutil.rmtree(output_path)
            args = ['termtosvg', 'render', cast_filename, output_path, '-s']
            TestMain.run_main(args, [])

        with self.subTest(case='record and render custom command'):
            args = ['termtosvg', '--command', 'ls']
            TestMain.run_main(args, [])

        with self.subTest(case='record and render on the fly (fallback theme)'):
            args = ['termtosvg', '--screen-geometry', '82x19']
            TestMain.run_main(args, SHELL_INPUT)

        with self.subTest(case='record and render on the fly (window_frame_js template)'):
            args = ['termtosvg', svg_filename, '--template', 'window_frame_js']
            TestMain.run_main(args, SHELL_INPUT)

        for template in termtosvg.config.default_templates():
            with self.subTest(case='record and render on the fly ({} template)'.format(template)):
                args = ['termtosvg', '-t', template]
                TestMain.run_main(args, SHELL_INPUT)

        cast_v1_data = '\r\n'.join([
            '{',
            '  "version": 1,',
            '  "width": 80,',
            '  "height": 32,',
            '  "duration": 10,',
            '  "command": "/bin/zsh",',
            '  "title": "",',
            '  "env": {},',
            '  "stdout": [',
            '    [0.010303, "\\u001b[1;31mnico \\u001b[0;34m~\\u001b[0m"],',
            '    [1.136094, "❤ ☀ ☆ ☂ ☻ ♞ ☯ ☭ ☢ € →"],',
            '    [0.853603, "\\r\\n"]',
            '  ]',
            '}',
        ])

        with self.subTest(case='render v1 cast file'):
            _, cast_filename_v1 = tempfile.mkstemp(prefix='termtosvg_', suffix='.cast')
            with open(cast_filename_v1, 'w') as cast_file:
                cast_file.write(cast_v1_data)

            args = ['termtosvg', 'render', cast_filename_v1, svg_filename]
            TestMain.run_main(args, [])

    def test_integral_duration(self):
        test_cases = [
            '100',
            '100ms',
            '100Ms',
        ]
        for case in test_cases:
            with self.subTest(case=case):
                self.assertEqual(
                    termtosvg.main.integral_duration_validation(case),
                    100
                )

